package custom;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.HashMap;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.avalon.framework.logger.ConsoleLogger;
import org.apache.avalon.framework.logger.Logger;
import org.apache.fop.apps.Driver;
import org.apache.fop.apps.FOPException;
import org.apache.fop.messaging.MessageHandler;
import org.w3c.tidy.Tidy;
import org.xml.sax.InputSource;

import com.jalios.jcms.Channel;
import com.jalios.jcms.JcmsConstants;
import com.jalios.jcms.JcmsUtil;
import com.jalios.jcms.Member;
import com.jalios.jcms.Publication;
import com.jalios.jcms.servlet.JcmsServlet;
import com.jalios.util.Util;


public class PDFServlet extends JcmsServlet {
  
  public void doGet (HttpServletRequest request, HttpServletResponse response) throws ServletException, java.io.IOException {
    
    String      xmlHtml      = request.getParameter("html");
    String      xmlJsp       = request.getParameter("jsp");
    String      xsl          = request.getParameter("xsl");
    Publication pub          = channel.getPublication(request.getParameter("id"));
    String      redirect     = Util.getString(request.getParameter("redirect"),"index.jsp");
    Member      loggedMember = channel.getAuthenticatedMemberFromCookies(request.getCookies());
    
    if (Util.isEmpty(xmlJsp) || Util.isEmpty(xmlHtml) || Util.isEmpty(xsl)){
      System.out.println("Missing Parameters");
      response.sendRedirect(redirect);
      return;
    }
    if (pub == null){
      System.out.println("Publication is Null");
      response.sendRedirect(redirect);
      return;
    }
    if (!pub.canBeReadBy(loggedMember)) {
      response.sendRedirect(JcmsConstants.FORBIDDEN_REDIRECT);
      return;
    }
    
    File xslFile = new File(channel.getRealPath(xsl));
    if (!xslFile.exists()){
      System.out.println("XSL not found");
      response.sendRedirect(redirect);
      return;
    }
    
    response.setContentType("application/pdf");
    response.setHeader("Content-Disposition", "attachment; filename="+pub.getTitle().replace(' ','_')+".pdf");
    
    try{
      InputStream stream = generateJSP(xmlJsp,channel,loggedMember,pub);
                  stream = Util.toBoolean(xmlHtml,false) ? generateXHTML(stream) : stream;
      convertXML2PDF(stream,new FileInputStream(xslFile),response.getOutputStream());
    }
    catch(Exception ex){ ex.printStackTrace(); throw new IOException(ex.getMessage()); }
  }
  
  
  public static synchronized InputStream generateXHTML(InputStream html) throws IOException{
  
      Tidy tidy = new Tidy();
      tidy.setXmlOut(true);
      tidy.setErrout(new PrintWriter(System.err));
      tidy.setMakeClean(true);
      tidy.setXHTML(true);

      ByteArrayOutputStream out = new ByteArrayOutputStream();
      tidy.parse(html, out);
      ByteArrayInputStream xhtml = new ByteArrayInputStream(out.toByteArray());
      
      html.close();
      out.close();
  
      return xhtml;
  }
  
  /**
   * Get the XML from a JSP processed by the template engine
   * @param jsp
   * @param channel
   * @param pub
   * @return InputStream
   * @throws IOException
   * @since jcms-4.0
   */
  public static synchronized InputStream generateJSP(String jsp, Channel channel, Member loggedMember, Publication pub) throws IOException{

    ByteArrayOutputStream out = new ByteArrayOutputStream();
    
    HashMap requestParams = new HashMap();
    requestParams.put("id",pub.getId());
      
    try{ JcmsUtil.jsp2OutputStream("/"+jsp,out,requestParams, null, null,null,null, channel,loggedMember,false); }
    catch(Exception ex){ ex.printStackTrace(); throw new IOException(ex.getMessage());  }
    
    ByteArrayInputStream in = new ByteArrayInputStream(out.toByteArray());
    out.close();

    return in;
  }
  
  /**
   * Convert the FO Stream to PDF Stream
   * @param fo
   * @param pdf
   * @throws IOException
   * @throws FOPException
   * @throws TransformerException
   * @since jcms-4.0
   */
  public static synchronized void convertFO2PDF(InputStream fo, OutputStream pdf) throws IOException, FOPException, TransformerException {
    //Construct driver
    Driver driver = new Driver();
        
    //Setup logger
    Logger logger = new ConsoleLogger(ConsoleLogger.LEVEL_INFO);
    driver.setLogger(logger);
    MessageHandler.setScreenLogger(logger);

    //Setup Renderer (output format)        
    driver.setRenderer(Driver.RENDER_PDF);
        
    //Setup output
    try {
        driver.setOutputStream(pdf);

        //Setup input
        try {
            driver.setInputSource(new InputSource(fo));
            
            //Process FO
            driver.run();
        } finally {
            fo.close();
        }
    } finally { pdf.close(); }    
  }
    
  /**
   * Convert the XML Stream to FO Stream using XSL as converter
   * @param xml
   * @param xsl
   * @param fo
   * @throws IOException
   * @throws FOPException
   * @throws TransformerException
   * @since jcms-4.0
   */
  public static synchronized  void convertXML2FO(InputStream xml, InputStream xsl, OutputStream fo) throws IOException, FOPException, TransformerException {
    //Setup output
    try {
        //Setup XSLT
        TransformerFactory factory = new org.apache.xalan.processor.TransformerFactoryImpl();
        Transformer transformer = factory.newTransformer(new StreamSource(xsl));
          
        
        //Setup input for XSLT transformation
        Source src = new StreamSource(xml);
        
        //Resulting SAX events (the generated FO) must be piped through to FOP
        Result res = new StreamResult(fo);

        //Start XSLT transformation and FOP processing
        transformer.transform(src, res);
    } finally { fo.close(); }    
  }

  /**
   * Convert the XML Stream to PDF Stream using XSL as converter
   * @param xml
   * @param xsl
   * @param pdf
   * @throws IOException
   * @throws FOPException
   * @throws TransformerException
   * @since jcms-4.0
   */
  public static synchronized  void convertXML2PDF(InputStream xml, InputStream xsl, OutputStream pdf) throws IOException, FOPException, TransformerException {
      //Construct driver
      Driver driver = new Driver();
        
      //Setup logger
      Logger logger = new ConsoleLogger(ConsoleLogger.LEVEL_INFO);
      driver.setLogger(logger);
      MessageHandler.setScreenLogger(logger);

      //Setup Renderer (output format)        
      driver.setRenderer(Driver.RENDER_PDF);

      //Setup output
      try {
          driver.setOutputStream(pdf);

          //Setup XSLT
          TransformerFactory factory = new org.apache.xalan.processor.TransformerFactoryImpl();
          Transformer transformer = factory.newTransformer(new StreamSource(xsl));
          
          //Setup input for XSLT transformation
          Source src = new StreamSource(xml);
        
          //Resulting SAX events (the generated FO) must be piped through to FOP
          Result res = new SAXResult(driver.getContentHandler());

          //Start XSLT transformation and FOP processing
          transformer.transform(src, res);
      } finally {
          pdf.close();
      }
  }
  
}
